import sys
import os

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from core.symbol import *

def is_sub (expr):
    if isinstance(expr, Mul) and expr.operands[0].is_negone():
        return True

def is_div (expr):
    if isinstance(expr, Pow) and expr.operands[1].is_negone():
        return True

def str_flat(expr):
    if expr.is_leaf():
        return str(expr)
    elif isinstance(expr, Add):
        result = ''
        for i, op in enumerate(expr.operands):
            if is_sub(op):
                result += str_flat(op) # - op. no + sign
            elif is_div(op):
                result += ' + 1' # will become + 1 / base
                result += str_flat(op)
            elif i == 0:
                result += str_flat(op) # no + sign in the first
            else:
                result += ' + ' #normal case
                result += str_flat(op)
        return result

    elif isinstance(expr, Mul):
        result = ''

        def helper(op):
            tempstr = str_flat(op)
            if isinstance(op, Add):
                tempstr = '(' + tempstr + ')'
            return tempstr

        for i, op in enumerate(expr.operands):
            if i == 0:
                if is_sub(expr):
                    result += ' - '
                elif is_div(op):
                    result += ' 1'
                    result += helper(op)
                else:
                    result += helper(op)
            elif i == 1:
                if is_sub(expr):
                    if is_div(op):
                        result += ' 1'
                    result += helper(op)
                else:
                    result += ' * '
                    result += helper(op)
            elif is_div (op):
                result += helper(op)
            else:
                result += ' * '
                result += helper(op)
        return result

    elif isinstance(expr, Pow):
        result = ''
        def helper(op):
            tempstr = str_flat(op)
            if isinstance(op, Add) or isinstance(op, Mul):
                tempstr = '(' + tempstr + ')'
            return tempstr

        base_str = helper(expr.base)
        exponent_str = helper(expr.exponent)
        if is_div(expr):
            result += ' / '
            result += base_str
        else:
            result += base_str
            result += ' ** '
            result += exponent_str
        return result






import sys
import os

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from core.symbol import *
from core.solve import Eq

# no sin cos tan log for now.
# all multiplications MUST be explicit use a * b to mean multiplication ab will be treated as one variable named ab


'''
prescedence ordering
1: binary +-
2: * /
3: unary +-
4: 
'''
# use Token for operators. Variables and Numbers get directly converted to exprs
class Token():
    def __init__(self,token_str:str, stack_level:int, unary = False):
        # +, -, *, /, **, var, num
        self.token_str = token_str
        self.unary = unary
        self.left_associative = True
        self.prescedence = 0

        if (unary == True or token_str in ('**')):
            self.left_associative = False

        if (unary == False and token_str in ('+','-')):
            self.prescedence += 1
        elif (token_str in ('*', '/')):
            self.prescedence += 2
        elif (token_str in ('**')):
            self.prescedence += 3
        self.prescedence += stack_level * 4

    # the idea is you would only consume operands with higher prescedence than yourself due to the conversation from high -> low prescedence. or it's a leaf with no prescedence.
    def to_expr(self, operands: list[Expr]):
        for op in operands:
            assert isinstance(op, Expr)
        if self.unary == True:
            assert(len(operands)) == 1
        else:
            assert(len(operands)) == 2

        if self.token_str == '+':
            return Add(*operands)
        elif self.token_str == '-' and self.unary == True:
            return Mul(-1, operands[0])
        elif self.token_str == '-' and self.unary == False:
            return Add(operands[0], Mul(-1,operands[1]))
        elif self.token_str == '*':
            return Mul(*operands)
        elif self.token_str == '/':
            return Mul(operands[0], Pow(operands[1],-1))
        elif self.token_str == '**':
            return Pow(operands[0], operands[1])

def _parse_exprstr_to_tokenlist(expr_str : str):
    token_list = []
    #for unary op classifying
    last_token_is_operator = True
    #for parenthesis
    stack_level = 0
    i = 0
    while i < len(expr_str):
        char = expr_str[i]
        if char == ' ' or char == ',':
            pass
        elif char == '(':
            stack_level += 1
        elif char == ')':
            stack_level -= 1
        elif char.isdigit() or char == '.':
            start = i
            while i < len(expr_str) and (expr_str[i].isdigit() or expr_str[i] == '.'):
                i += 1
            number_str = expr_str[start:i]
            token_list.append(Number(float(number_str)))
            i -= 1
            last_token_is_operator = False
        elif char.isalpha():
            start = i
            while i < len(expr_str) and (expr_str[i].isalpha() or expr_str[i] == '_'):
                i += 1
            variable_str = expr_str[start:i]
            token_list.append(Variable(variable_str))
            i -= 1
            last_token_is_operator = False
        elif char in ('+', '-'):
            if last_token_is_operator:
                token_list.append(Token(char, stack_level,True))
            else:
                token_list.append(Token(char, stack_level,False))
            last_token_is_operator = True
        elif char == '/':
            token_list.append(Token(char, stack_level))
            last_token_is_operator = True
        elif char == '*':
            # handle the ** case
            if i + 1 < len(expr_str) and expr_str[i + 1] == '*':
                token_list.append(Token('**', stack_level))
                i += 1  # Skip the next '*'
            else:
                token_list.append(Token('*', stack_level))
            last_token_is_operator = True           
        else:
            raise ValueError('string not supported cannot parse')
        i += 1
    return token_list

def _create_AST_from_tokenlist(token_list):
    prescedence_ordering = set()
    for token in token_list:
        if isinstance(token, Token):
            prescedence_ordering.add(token.prescedence)
    sorted_ordering = sorted(prescedence_ordering, reverse = True)
    for cur_prescedence in sorted_ordering:
        i = 0
        while i < len(token_list):
            cur_token = token_list[i]
            if (isinstance(cur_token, Token) and
                cur_token.prescedence == cur_prescedence and
                cur_token.left_associative == True):
                # no unary ops should be left associtive.
                assert(cur_token.unary == False)
                cur_expr = cur_token.to_expr([token_list[i-1], token_list[i+1]])
                token_list[i-1 : i+2] = [cur_expr]
                i -= 1
            i += 1

        i = len(token_list) - 1
        while i >= 0:
            cur_token = token_list[i]
            if (isinstance(cur_token, Token) and
                cur_token.prescedence == cur_prescedence and
                cur_token.left_associative == False):

                print(cur_token)

                if cur_token.unary == True:
                    cur_expr = cur_token.to_expr([token_list[i+1]])
                    token_list[i : i+2] = [cur_expr]
                elif cur_token.unary == False:
                    cur_expr = cur_token.to_expr([token_list[i-1],token_list[i+1]])
                    token_list[i-1 : i+2] = [cur_expr]
                    i -= 1
            i -= 1
    assert (len(token_list) == 1)
    assert (isinstance(token_list[0], Expr))
    return token_list[0]

# handles = to create Eq here
def parse_expr(expr_str: str):
    for i, char in enumerate(expr_str):
        if char == '=':
            left_part = expr_str[:i]
            right_part = expr_str[i+1:]
            left_expr = parse_expr(left_part)
            right_expr = parse_expr(right_part)
            return Eq(left_expr, right_expr)
    
    token_list = _parse_exprstr_to_tokenlist(expr_str)
    expr = _create_AST_from_tokenlist(token_list)
    return expr





# Expr is immutable

import math
from typing import Tuple


class Expr:

    def print_tree(self, indent = 0):
        print(' ' * indent + "└── " + str(self))
        if not self.is_leaf():
            for operand in self.operands:
                operand.print_tree(indent + 4)

    # TODO: prettier print no * /c things like this just do / c
    def print_flat(self):
        def str_flat(expr):
            # Special case: -1 * a becomes -a
            if (isinstance(expr, Mul) and 
                len(expr.operands) == 2 and
                isinstance(expr.operands[0], Number) and
                expr.operands[0].value == -1):
                return "-" + str_flat(expr.operands[1])
            
            # Special case: a ^ -1 becomes /a  
            elif (isinstance(expr, Pow) and
                  isinstance(expr.exponent, Number) and
                  expr.exponent.value == -1):
                return "/" + str_flat(expr.base)
            if expr.is_leaf():
                return str(expr)
            else:
                operands = [str_flat(op) for op in expr.operands]
                return "(" + f" {str(expr)} ".join(operands) + ")"
        print(str_flat(self))


    # arithmatic unary
    def __neg__(self): return Mul(-1, self)
    def __pos__(self): return Mul(1, self)
    # arithmatic binary
    def __add__ (self, other): return Add(self, other)   
    def __radd__ (self,other): return Add(other, self)
    def __sub__(self, other): return Add(self,Mul(-1,other))
    def __rsub__(self, other): return Add(other,Mul(-1,self))
    def __mul__(self, other): return Mul(self,other)
    def __rmul__(self, other): return Mul(other,self)   
    def __truediv__(self, other): return Mul(self,Pow(other,-1))
    def __rtruediv__(self, other): return Mul(other,Pow(self,-1))
    def __pow__(self, other):   return Pow(self,other)
    def __rpow__(self, other): return Pow(other,self)
    # comparison binary
    def __eq__(self,other): return Compare('==',self,other)
    def __ne__(self,other): return Compare('!=',self,other)
    def __lt__(self,other): return Compare('<',self,other)
    def __le__(self,other): return Compare('<=',self,other)
    def __gt__(self,other): return Compare('>',self,other)
    def __ge__(self,other): return Compare('>=',self,other)
    # logical unary
    def __invert__(self): return Logic('~', self)
    # logical variadic
    def __and__(self,other): return Logic('&',self,other)
    def __or__(self,other): return Logic('|',self,other)
    def __xor__(self,other): return Logic('^',self,other)

    #helper functions
    def is_leaf(self): return isinstance(self, Number) or isinstance(self, Variable)
    def is_number(self): return isinstance(self,Number)
    def is_zero(self): return isinstance(self, Number) and abs(self.value) < 1e-12
    def is_one(self): return isinstance(self, Number) and abs(self.value - 1.0) < 1e-12
    def is_negone(self): return isinstance(self, Number) and abs(self.value + 1.0) < 1e-12

    def _as_add_terms(self):
        if isinstance(self, Add):
            return self.operands
        else:
            return (self,)

    def _as_mul_factors(self):
        if isinstance(self, Mul):
            return self.operands
        else:
            return (self,)

    def _as_coef_literal(self):
        if isinstance(self, Mul):
            coefficient = Number(1) 
            ops = []
            for operand in self.operands:
                if isinstance(operand, Number):
                    coefficient = operand
                else:
                    ops.append(operand)
            return coefficient, Mul(*ops)
        return Number(1), self

    def _as_base_exp(self):
        if isinstance(self,Pow):
            return self.operands[0], self.operands[1]
        return self, Number(1)

    '''
    def contains_var(self,var:Variable) -> bool:
        if self.is_leaf():
            if self == var: return True
        else:
            for op in self.operands:
                if contains_var(op):
                    return True
        return False
    '''


    #hash and comparison (same logic)
    def __hash__(self): return hash(canon_key(self))
    def __eq__(self, other):
        if not isinstance(other,Expr):
            return False
        return canon_key(self) == canon_key(other)

    # expand 
    def expand_tree(self):
        if self.is_leaf():
            return self
        else:
            new_operands = []
            for operand in self.operands:
                new_operand = operand.expand_tree()
                if new_operand is not None:
                    new_operands.append(new_operand)

            new = type(self)(*new_operands)
            new = new.expand()
            return new
    
    def expand(self):
        return self

def _ensure_expr(value):
    if isinstance(value,(int, float)):
        return Number(value)
    assert isinstance(value,Expr)
    return value


# does not deal with logical ops for now
# communicative ops' operands are sorted non communicative ops' operand are kept in order
def canon_key(expr):
    if isinstance(expr, Number): return ('0NUM', str(expr)) # put 0 in name just for it be in front for printing clarity
    elif isinstance(expr, Variable): return ('1VAR', str(expr))
    elif isinstance(expr, Add): return ('2ADD', tuple(sorted(canon_key(o) for o in expr.operands)))
    elif isinstance(expr, Mul): return ('3MUL', tuple(sorted(canon_key(o) for o in expr.operands)))
    elif isinstance(expr, Pow): return ('4POW', tuple(canon_key(o) for o in expr.operands))

    
class Number(Expr):
    def __init__(self, value):
        assert isinstance(value,(int,float))
        self.value = float(value)

    def __str__(self):
        return f"{self.value:.{12}g}"
    
class Boolean(Expr):
    def __init__(self, value):
        assert value in (True,False)
        self.value = value

    def __str__(self):
        return str(self.value)

class Variable(Expr):
    def __init__(self, name):
        self.name = name
    
    def __str__(self):
        return str(self.name)

# local rules: add Numbers. combine same MONOMIAL terms and add their COEFFICIENTS
# special rules. if total is 0 return 0. 
class Add (Expr):
    def __new__(cls, *args):

        constant = 0
        ops = []
        def merge (expr):
            nonlocal constant
            nonlocal ops
            expr = _ensure_expr(expr)
            # add constants directly
            if isinstance(expr,Number):
                constant += expr.value
            else:
                ops.append(expr)
        for arg in args:
            if isinstance(arg, Add):
                for operand in arg.operands:
                    merge(operand)
            else:
                merge(arg)
        constant = Number(constant)
        # combine monomial terms that have the same literals and add their coefficients
        literal_to_coefs = {}
        for op in ops:
            coef, literal = op._as_coef_literal()

            if literal in literal_to_coefs:
                literal_to_coefs[literal] = literal_to_coefs[literal] + coef
            else:
                literal_to_coefs[literal] = coef

        ops = []
        for literal, coef in literal_to_coefs.items():
            if coef.is_zero() == False:
                #ops.append(coef * literal)
                temp = coef * literal
                merge(temp)

        # if constant is not zero add it to ops
        if not constant.is_zero():
            ops.append(constant)
        # if everything is constant(no ops) just return a number
        if len(ops) == 0:
            return constant
        # if it's just a unary op directly return
        if len(ops) == 1:
            return ops[0]
        ops.sort(key = canon_key)
        add = object.__new__(cls)
        add.operands = tuple(ops)
        return add

    def __str__(self):
        return '+'
        

# local rules: multiply Numbers. combine same BASES add their EXPONENTS
# special rules: if coefficient is 0, return 0. if coefficient is 1, delete coefficient.
class Mul (Expr):
    def __new__ (cls, *args):
        coefficient = 1
        ops = []
        def merge (expr):
            nonlocal coefficient
            nonlocal ops
            expr = _ensure_expr(expr)
            # multiply numbers directly.
            if isinstance(expr, Number):
                coefficient *= expr.value
            else:
                ops.append(expr)
        for arg in args:
            if isinstance(arg, Mul):
                for operand in arg.operands:
                    merge(operand)
            else:
                merge(arg)
        coefficient = Number(coefficient)
        # if coefficient is 0 return 0
        if coefficient.is_zero():
            return Number(0)
        # combine same bases add exponents
        base_to_exps = {}
        for op in ops:
            base, exp = op._as_base_exp()
            if base in base_to_exps:
                base_to_exps[base] = base_to_exps[base] + exp
            else:
                base_to_exps[base] = exp
        ops = []
        for base, exp in base_to_exps.items():
            ops.append(base**exp)

        # if coefficient is 1 delete coefficient
        if not coefficient.is_one():
            ops.append(coefficient)
        # if only one factor just return that
        if len(ops) == 1:
            return ops[0]
        ops.sort(key = canon_key)
        mul =  object.__new__(cls)
        mul.operands = tuple(ops)
        return mul 

    def __str__(self):
        return '*'
    
    # distribute over add operands. i.e. 3*(a+b) -> 3*a+3*b
    def expand(self) -> Add:
        plus_terms:List[Add] = []
        combine_terms:List[Expr] = []
        for operand in self.operands:
            if isinstance(operand, Add):
                plus_terms.append(operand)
            else:
                combine_terms.append(operand)

        combine_term = Mul(*combine_terms)

        for plus_term in plus_terms:
            new_term = Number(0)
            for plus_op in plus_term.operands:
                new_term = new_term + plus_op * combine_term
            combine_term = new_term

        return combine_term


#NOTE: maybe also make this a chain but an associative order matters chain?
class Pow (Expr):
    def __new__ (cls, base, exponent):
        base = _ensure_expr(base)
        exponent = _ensure_expr(exponent)
        # special rules for when bases or power is zero or one
        if base.is_zero():
            if isinstance(exponent, Number):
                assert exponent.value > 0
            return Number(1)
        elif base.is_one():
            return Number(1)
        elif exponent.is_zero():
            return Number(1)
        elif exponent.is_one():
            return base
        elif isinstance(base, Number) and isinstance(exponent, Number):
            return Number(base.value ** exponent.value)
        # if base is pow have exponents changed to multiplications
        if isinstance(base, Pow):
            inner_base = base.base
            inner_exponent = base.exponent
            new_exponent = exponent * inner_exponent
            base = inner_base
            exponent = new_exponent

        pow = object.__new__(cls)
        pow.base = base
        pow.exponent = exponent
        pow.operands = (base,exponent)
        return pow

    def __str__(self):
        return '**'

class Compare(Expr):
    def __init__(self, operator, operand_A: Expr, operand_B: Expr):
        assert operator in {'==','!=','<','<=','>','>='} 
        self.operator = operator
        self.operands = (_ensure_expr(operand_A),_ensure_expr(operand_B))

    def __str__(self):
        return self.operator

class Logic(Expr):
    def __init__(self, operator, *args):
        assert operator in {'&','|','^'}                  
        self.operator = operator
        # FIXME: this is wrong perhaps. need to extend things if its a logic of the same operator but fix later
        ops = [_ensure_expr(arg) for arg in args]
        ops.sort(key = canon_key)
        self.operator = tuple(ops) 

    def __str__(self):
        return self.operator

class If(Expr):
    def __init__(self, condition: Expr, if_branch: Expr, else_branch: Expr):
        super().__init__()
        self.type = 'If'
        self.condition = condition
        self.if_branch = if_branch
        self.else_branch = else_branch
        condition.laters.append(self)
        if_branch.laters.append(self)
        else_branch.laters.append(self)

    def print_tree(self, indent = 0):
        print(' ' * indent + "└── " + self.type)
        self.condition.print_tree(indent + 4)
        self.if_branch.print_tree(indent + 4)
        self.else_branch.print_tree(indent + 4)


# class Derivative(Expr):

# class Integral(Expr):

# arithmatic unary functions
def sin(x): return UnaryOp('sin', _ensure_expr(x))
def cos(x): return UnaryOp('cos', _ensure_expr(x))
def tan(x): return UnaryOp('tan', _ensure_expr(x))
# arithmatic binary functions
def log(x,base): return BinaryOp('log', _ensure_expr(x),_ensure_expr(base))


'''
to define complex user functions just use python's normal syntax

i.e.
def foo (x,y,z)
    a = x + y * z
    b = 3 - x ** y
    return a, b
i.e.
def relu(x):
    result = If(x>0, x, Number(0))
    return result
'''






from setuptools import setup, find_packages
setup(
    name="svmbolcore",
    version="0.1.11",
    author="codespaceDrifter",  
    author_email="codespaceDrifter@gmail.com",  
    description="symbolic math lib that generates parallel CUDA",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/yourusername/svmbolcore",  # Your GitHub URL
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
    ],
)



'''
how to update note to self:
-> Update version in setup.py (e.g. "0.1.0" -> "0.1.1")
commands:
(run in root folder where setup.py is)
python -m build
python -m twine upload dist/*
'''



import enum
from .symbol import *
from .symbol import _ensure_expr

#returns an expression that all solvers assume equal to zero
class Eq(Expr):
    def __init__(self, left: Expr, right:Expr = Number(0)):
        self.left = _ensure_expr(left)
        self.right = _ensure_expr(right)
        self.operands = (left,right)

    def __str__(self):
        return '='
    
    def to_expr(self) -> Expr:
        new_left = self.left - self.right
        return new_left



# assumes expression is simplified
def get_degree(expr, vars:[list]):
    if isinstance(expr, Number):
        return 0
    elif expr in vars:
        return 1
    elif isinstance(expr, Pow) and expr.base in vars:
        assert isinstance(expr.exponent, Number), "Complex variable exponents not supported yet"
        return int(expr.exponent.value)
    elif isinstance(expr, Mul):
        total_degree = 0
        for operand in expr.operands:
            total_degree += get_degree(operand, vars)
        return total_degree
    elif isinstance(expr, Add):
        max_degree = 0
        for operand in expr.operands:
            operand_degree = get_degree(operand, vars)
            if operand_degree > max_degree:
                max_degree = operand_degree
        return max_degree
    else:
        return 0

# no self definition or recurring definitions. i.e. a cant be defined with b and then b defined with a. 
def assert_solutions_valid (solutions: dict[Variable, Expr]):
    # get what variables each solution contains
    def extract_variables(expr):
        contained_vars = set()
        def helper(expr):
            nonlocal contained_vars
            if isinstance(expr, Variable) and expr not in contained_vars:
                contained_vars.add(expr)
            if not expr.is_leaf():
                for op in expr.operands:
                    helper(op)
        helper(expr)
        return contained_vars
    # put solutions into a variable : the variables it's definition contains form
    var_to_defvars = {}
    for var, expr in solutions.items():
        var_to_defvars[var] = extract_variables(expr)

    # while deep traversing along one branch mark as visiting, while finished traversing mark as visited on the function end back up
    # because A->B, and A->D->B is fine but A->B->C and then A->C->B is not (C will lead to B before the recursion calls end and B will still be in visiting so it's A->B->C->B)
    visiting = set()
    visited = set()
    def dfs(var):
        if var in visiting:
            return True
        if var in visited:
            return False
        visiting.add(var)
        for defvar in var_to_defvars[var]:
            if defvar in var_to_defvars:
                if dfs(defvar):
                    return True
        visiting.remove(var)
        visited.add(var)
        return False

    for var in var_to_defvars:
        if var not in visited and dfs(var):
            raise ValueError('cycle definitions')

def subsitute (expr, solutions: dict[Variable,Expr]):
    assert_solutions_valid(solutions)
    def helper(expr):
        if expr in solutions:
            return solutions[expr]
        elif expr.is_leaf():
            return expr
        else:
            new_operands = []
            for op in expr.operands:
                new_op = helper(op)
                new_operands.append(new_op)
            return type(expr)(*new_operands)
    cur = expr
    while True:
        next = helper(cur)
        if next == cur:
            break
        cur = next
    return cur

def print_solutions(sol):
    for key,value in sol.items():
        print('variable')
        key.print_flat()
        print('solution')
        value.print_flat()



def solve(equation_list:list[Expr], vars:list[Variable]):
    expr_list = []
    for e in equation_list:
        if isinstance(e, Eq):
            expr_list.append(e.to_expr())
        else:
            expr_list.append(e)
    # get max degree of all expressions
    max_degree = 0
    for expr in expr_list:
        cur_degree = get_degree(expr, vars)
        if cur_degree > max_degree: max_degree = cur_degree
    # send to different solvers
    if max_degree == 1:
        return solve_linear_system(expr_list, vars)

def solve_linear_system(exprs: list[Expr], vars: list[Variable])->dict[Variable,Expr]:
    num_exprs = len(exprs)
    num_vars = len(vars)
    assert (num_exprs == num_vars)
    # extract parameters and numbers coefficients for a specific variable
    #assume the expr is one degree max. therefore no pow. 
    def extract_parameter_coef(expr, var):
        terms = expr._as_add_terms()
        coefs = Number(0)
        for term in terms:
            factors = term._as_mul_factors()
            coef = Number(1)
            hasVar = False
            for factor in factors: # this case does not have var as the base of a power
                if factor == var:
                    hasVar = True
                else:
                    coef = coef * factor
            if hasVar:
                coefs = coefs + coef
        return coefs
    # gets the parameters and constant and negates it
    def extract_neg_constant_parameters(expr, vars):
        terms = expr._as_add_terms()
        neg = Number(0)
        for term in terms:
            factors = term._as_mul_factors()
            value = Number(1)
            hasVar = False

            for factor in factors:
                if factor in vars:
                    hasVar = True
                    break
                else:
                    value = value * factor
            if not hasVar:
                neg = neg - value
        return neg
    # augmented matrix of Ax = -b. constant is negative
    matrix = []
    for expr in exprs:
        row = []
        for var in vars:
            row.append(extract_parameter_coef(expr, var))
        row.append(extract_neg_constant_parameters(expr, vars))
        matrix.append(row)


    def find_pivot_row(matrix, col):
        num_rows = len(matrix)
        for i in range (col, num_rows):
            if not matrix[i][col].is_zero():
                return i
        raise ValueError('infinite solutions not supported yet')
    for i in range(num_exprs):
        pivot_row = find_pivot_row(matrix, i)
        matrix[i], matrix[pivot_row] = matrix[pivot_row], matrix[i]
        pivot = matrix[i][i]
        assert not pivot.is_zero()
        for j in range (i+1, num_exprs):
            factor = matrix[j][i] / pivot
            for k in range (i, num_exprs + 1): # because one additional const column
                matrix[j][k] = matrix[j][k] - matrix[i][k] * factor
    solutions = {}
    # start at the last row 2nd last col. go diagnolly upwards
    # minus augmented row with coef * solution[var] of latter cols
    for i in range (num_exprs -1, -1, -1):
        value = matrix[i][-1]
        for j in range (num_exprs - 1, i, -1): #2nd last col to after current col
            coef_solution = matrix[i][j]
            literal_solution = solutions[vars[j]]
            value = value - coef_solution * literal_solution
        coef = matrix[i][i]
        sol = value / coef
        solutions[vars[i]] = sol
    return solutions


#TODO: do NOT do quadratic system. just do single equation. no time for now. 
def solve_quadratic(expr, var)-> Number:
    pass

